"""Provides the Markdown viewer widget."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from functools import singledispatchmethod
from os import getenv
from pathlib import Path
from subprocess import run

##############################################################################
# httpx imports.
from httpx import URL, AsyncClient, HTTPStatusError, RequestError

##############################################################################
# MarkdownIt imports.
from markdown_it import MarkdownIt
from mdit_py_plugins import front_matter

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Vertical, VerticalScroll
from textual.events import Click
from textual.message import Message
from textual.reactive import var
from textual.widgets import Label, Markdown, Rule

##############################################################################
# Local imports.
from .. import USER_AGENT
from ..commands import JumpToCommandLine
from ..data import is_editable, load_configuration, looks_urllike
from ..editor import Editor
from ..messages import CopyToClipboard, OpenLocation
from ..support import is_copy_request_click, view_in_browser
from ..types import HikeHistory, HikeLocation


##############################################################################
class ViewerTitle(Label):
    """Widget to display the viewer's title."""

    DEFAULT_CSS = """
    ViewerTitle {
        background: $panel;
        color: $foreground;
        content-align: right middle;
        width: 1fr;
        height: 1;
    }
    """

    location: var[HikeLocation | None] = var(None, always_update=True)
    """The location to display."""

    def _watch_location(self) -> None:
        """React to the location changing."""
        if (
            len(
                display := ""
                if self.location is None
                else str(self.location)[-self.size.width :]
            )
            >= self.size.width
        ):
            display = f"…{display[1:]}"
        self.update(display)

    def on_resize(self) -> None:
        """Handle the widget being resized."""
        self.location = self.location

    @on(Click)
    def _maybe_clipboard(self, message: Click) -> None:
        """Maybe copy to the clipboard.

        Args:
            message: The mouse click message.
        """
        if is_copy_request_click(message) and self.location:
            message.stop()
            self.post_message(CopyToClipboard(str(self.location)))


##############################################################################
class Viewer(Vertical, can_focus=False):
    """The Markdown viewer widget."""

    DEFAULT_CSS = """
    Viewer {
        display: block;
        &.empty {
            display: none;
        }
        &> VerticalScroll {
            background: transparent;
        }
        Markdown {
            background: transparent;
        }
        Rule {
            height: 1;
            margin: 0 !important;
            color: $foreground 10%;
        }
    }
    """

    HELP = """
    ## Viewer

    This is the main Markdown viewer.

    ### Copying to the clipboard

    See the main help and the command palette for commands for copying to
    the clipboard; you can also copy with the mouse by either clicking once
    while holding down <kbd>ctrl</kbd>, or click 3 times. Doing so on the
    location will copy the location, doing so on the main document will copy
    the markdown's source.
    """

    BINDINGS = [("escape", "bounce_out")]

    location: var[HikeLocation | None] = var(None)
    """The location of the markdown being displayed."""

    history: var[HikeHistory] = var(HikeHistory)
    """The history for the viewer."""

    _source: var[str] = var("")
    """The source of the Markdown we're viewing."""

    def compose(self) -> ComposeResult:
        """Compose the content of the viewer."""
        yield ViewerTitle()
        yield Rule(line_style="heavy")
        with VerticalScroll(id="document"):
            yield Markdown(
                open_links=False,
                parser_factory=lambda: MarkdownIt("gfm-like").use(
                    front_matter.front_matter_plugin
                ),
            )

    @property
    def source(self) -> str:
        """The source of the markdown being viewed."""
        return self._source

    @property
    def filename(self) -> Path | None:
        """The name of the file being viewed.

        Notes:
            This is just the name of the file itself, without any path. If
            no file is being viewed at the moment then the value is `None`.
        """
        if isinstance(self.location, Path):
            return Path(self.location.name)
        if isinstance(self.location, URL):
            return Path(Path(self.location.path).name)
        return None

    def action_bounce_out(self) -> None:
        """Bounce back out to the input."""
        self.post_message(JumpToCommandLine())

    @dataclass
    class Loaded(Message):
        """Class posted when the markdown content is loaded."""

        viewer: Viewer
        """The viewer."""

        markdown: str
        """The markdown content."""

        remember: bool
        """Should this load be remembered?"""

    @dataclass
    class HistoryUpdated(Message):
        """Class posted when the history is updated."""

        viewer: Viewer
        """The viewer."""

    @dataclass
    class HistoryVisit(Message):
        """Class posted when a location in history is visited."""

        viewer: Viewer
        """The viewer."""

    @on(HistoryUpdated)
    def _history_updated(self) -> None:
        """React to the bindings being updated."""
        self.refresh_bindings()
        # Given that the content of history has changed, we need to update
        # the display. If we have history...
        if self.history:
            # ...visit whatever is now current.
            self._visit_from_history()
        else:
            # ...otherwise there's nothing to display.
            self.location = None

    def _watch_history(self) -> None:
        """React to the history being updated."""
        self.post_message(self.HistoryUpdated(self))

    @work(thread=True, exclusive=True)
    def _load_from_file(self, location: Path, remember: bool) -> None:
        """Load up markdown content from a file.

        Args:
            location: The path to load the content from.
            remember: Should this location go into history?
        """
        try:
            self.post_message(
                self.Loaded(self, Path(location).read_text(encoding="utf-8"), remember)
            )
        except OSError as error:
            self.notify(str(error), title="Load error", severity="error", timeout=8)

    @work(exclusive=True)
    async def _load_from_url(self, location: URL, remember: bool) -> None:
        """Load up markdown content from a URL.

        Args:
            location: The URL to load the content from.
            remember: Should this location go into history?
        """

        # Download the data from the remote location.
        try:
            async with AsyncClient() as client:
                response = await client.get(
                    location,
                    follow_redirects=True,
                    headers={"user-agent": USER_AGENT},
                )
        except RequestError as error:
            self.notify(str(error), title="Request error", severity="error", timeout=8)
            return

        # We got a response, left's check it's a good one.
        try:
            response.raise_for_status()
        except HTTPStatusError as error:
            self.notify(str(error), title="Response error", severity="error", timeout=8)
            return

        # At this point we've got a good response. Now let's be sure that
        # what we got back is something that admits to being markdown, or at
        # least a form of plain text we can render.
        if content_type := response.headers.get("content-type"):
            if any(
                content_type.startswith(allowed_type)
                for allowed_type in load_configuration().markdown_content_types
            ):
                self.post_message(self.Loaded(self, response.text, remember))
                return

        # It doesn't look like Markdown, so let's open it in the browser.
        self.notify(
            "That location doesn't look like a Markdown file, opening in your browser..."
        )
        view_in_browser(location)

    @singledispatchmethod
    def _load_markdown(self, location: Path, remember: bool) -> None:
        """Load markdown from a location.

        Args:
            location: The location to load the markdown from.
            remember: Should this location go into history?
        """
        self._load_from_file(location, remember)

    @_load_markdown.register
    def _(self, location: URL, remember: bool) -> None:
        self._load_from_url(location, remember)

    @_load_markdown.register
    def _(self, location: None, remember: bool) -> None:
        self.post_message(self.Loaded(self, "", remember))

    def _visit(
        self,
        location: HikeLocation | None,
        remember: bool = True,
        preserve_position: bool = False,
    ) -> None:
        """Visit the given location.

        Args:
            location: The location to visit.
            remember: Should this location go into history?
            preserve_position: Attempt to preserve the scroll position?
        """
        self.set_class(location is None, "empty")
        self._load_markdown(location, remember)
        if not preserve_position:
            self.query_one("#document").scroll_home(animate=False)

    def _watch_location(self) -> None:
        """Handle changes to the location to view."""
        if self.is_mounted:
            self._visit(self.location)

    @on(Loaded)
    def _update_markdown(self, message: Loaded) -> None:
        """Update the markdown once some new content is loaded.

        Args:
            message: The message requesting the update.
        """
        self.query_one(ViewerTitle).location = self.location
        self._source = message.markdown
        self.query_one(Markdown).update(message.markdown)
        if (
            message.remember
            and self.location
            and self.location != self.history.current_item
        ):
            self.history.add(self.location)
            self.post_message(self.HistoryUpdated(self))

    def _visit_from_history(self) -> None:
        """Visit the current location in history."""
        self.set_reactive(Viewer.location, self.history.current_item)
        self._visit(self.location, remember=False)
        self.post_message(self.HistoryVisit(self))
        self.refresh_bindings()

    def reload(self) -> None:
        """Reload the current document."""
        self._visit(self.location, remember=False, preserve_position=True)

    def goto(self, history_location: int) -> None:
        """Go to a specific location in history."""
        if self.history.current_location != history_location:
            self.history.goto(history_location)
            self._visit_from_history()

    def backward(self) -> None:
        """Go backward through the history."""
        if self.history.backward():
            self._visit_from_history()

    def forward(self) -> None:
        """Go forward through the history."""
        if self.history.forward():
            self._visit_from_history()

    def jump_to_content(self, block_id: str) -> None:
        """Jump to some content in the current document.

        Args:
            block_id: The ID of the content to jump to.
        """
        self.scroll_to_widget(self.query_one(f"#{block_id}"), top=True)

    def remove_from_history(self, history: int) -> None:
        """Remove a specific location from history.

        Args:
            history: The ID of the location in history to remove.
        """
        del self.history[history]
        self.post_message(self.HistoryUpdated(self))

    def clear_history(self) -> None:
        """Clear all locations from history."""
        self.history = HikeHistory()

    @on(Markdown.LinkClicked)
    def _handle_link(self, message: Markdown.LinkClicked) -> None:
        """Handle a link being clicked in the Markdown widget.

        Args:
            message: The message requesting the link be handled.
        """
        message.stop()

        # Outright URL?
        if looks_urllike(message.href):
            self.post_message(OpenLocation(URL(message.href)))
            return

        # Possibly a relative path to a currently-visited URL?
        if isinstance(self.location, URL):
            self.post_message(
                OpenLocation(self.location.copy_with().join(message.href))
            )
            return

        # A local file that exists?
        if (local_file := Path(message.href).expanduser()).exists():
            self.post_message(OpenLocation(local_file.resolve()))
            return

        # A local file relative to the current location?
        if (
            isinstance(self.location, Path)
            and (local_file := self.location.parent / Path(message.href))
            .absolute()
            .exists()
        ):
            self.post_message(OpenLocation(local_file))
            return

        # Some sort of internal anchor perhaps?
        if message.href.startswith("#") and message.markdown.goto_anchor(
            message.href[1:]
        ):
            return

        self.notify(
            f"The clicked link could not be handled:\n\n{message.href}",
            title="Unknown link type",
            severity="error",
        )

    @on(Click)
    def _maybe_clipboard(self, message: Click) -> None:
        """Maybe copy to the clipboard.

        Args:
            message: The mouse click message.
        """
        if is_copy_request_click(message) and self._source:
            message.stop()
            self.post_message(CopyToClipboard(self._source))

    @property
    def is_editable(self) -> bool:
        """Is the current location editable?"""
        return self.location is not None and is_editable(self.location)

    def edit(self) -> None:
        """Edit the current document."""

        # There's no point in even attempting to edit if we're not looking a
        # something.
        if self.location is None:
            return

        # We can't edit something that isn't local.
        if not is_editable(self.location):
            self.notify(
                "Editing is only supported for Markdown files in the local filesystem.",
                title="Not Supported",
                severity="warning",
            )
            return

        # We need an editor to edit things.
        if editor := (getenv("VISUAL") or getenv("EDITOR") or ""):
            # Run the editor.
            with self.app.suspend():
                run((editor, str(self.location)))
                # Work around https://github.com/Textualize/textual/issues/5528.
                self.app.refresh()
            # Given we did an edit, we should now reload.
            self.reload()
        else:
            self.app.push_screen(
                Editor(self.location), callback=lambda _: self.reload()
            )


### viewer.py ends here



"""Tests for the command line command matching code."""

##############################################################################
# Pytest imports.
from pytest import mark

##############################################################################
# Local imports.
from hike.widgets.command_line.base_command import InputCommand


##############################################################################
class ExampleCommand(InputCommand):
    COMMAND = "`test`"
    ALIASES = "`t`, `tester`, `testing`"


##############################################################################
@mark.parametrize(
    "look_for, found",
    (
        ("test", True),
        ("t", True),
        ("tester", True),
        ("testing", True),
        ("TEST", True),
        ("T", True),
        ("TESTER", True),
        ("TESTING", True),
        ("  TeST  ", True),
        ("  T  ", True),
        ("  TEstER  ", True),
        ("  TESting  ", True),
        ("", False),
        ("te", False),
    ),
)
def test_is_command(look_for: str, found: bool) -> None:
    """We should be able to find if a command is a match."""
    assert ExampleCommand.is_command(look_for) is found


##############################################################################
@mark.parametrize(
    "split, result",
    (
        ("", ("", "")),
        ("test", ("test", "")),
        (" test ", ("test", "")),
        ("test test", ("test", "test")),
        (" test  test ", ("test", "test")),
        ("test test test test", ("test", "test test test")),
    ),
)
def test_split_command(split: str, result: tuple[str, str]) -> None:
    """The command splitting code should work as expected."""
    assert ExampleCommand.split_command(split) == result


### test_command_line_commands.py ends here



"""The main entry point for the application."""

##############################################################################
# Python imports.
from argparse import ArgumentParser, Namespace

##############################################################################
# Local imports.
from . import __version__
from .hike import Hike


##############################################################################
def get_args() -> Namespace:
    """Get the command line arguments.

    Returns:
        The arguments.
    """

    # Build the parser.
    parser = ArgumentParser(
        prog="Hike",
        description="Hike - A Markdown browser for the terminal",
        epilog=f"v{__version__}",
    )

    # Add --version
    parser.add_argument(
        "-v",
        "--version",
        help="Show version information",
        action="version",
        version=f"%(prog)s v{__version__}",
    )

    # The remainder is going to be the initial command.
    parser.add_argument(
        "command",
        help=("The initial command; " "can be any valid input to Hike's command line."),
        nargs="*",
    )

    # Finally, parse the command line.
    return parser.parse_args()


##############################################################################
def main() -> None:
    """The main entry point."""
    Hike(get_args()).run()


##############################################################################
if __name__ == "__main__":
    main()


### __main__.py ends here



"""Provides the main application commands for the command palette."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import (
    ChangeTheme,
    Command,
    CommandHits,
    CommandsProvider,
    Help,
    Quit,
)

##############################################################################
# Local imports.
from ..commands import (
    Backward,
    BookmarkLocation,
    ChangeCommandLineLocation,
    ChangeNavigationSide,
    CopyLocationToClipboard,
    CopyMarkdownToClipboard,
    Edit,
    Forward,
    JumpToBookmarks,
    JumpToCommandLine,
    JumpToHistory,
    JumpToLocalBrowser,
    JumpToTableOfContents,
    Reload,
    SaveCopy,
    SearchBookmarks,
    ToggleNavigation,
)


##############################################################################
class MainCommands(CommandsProvider):
    """Provides some top-level commands for the application."""

    def _maybe(self, command: type[Command]) -> CommandHits:
        """Yield a command if it's application.

        Args:
            command: The type of the command to maybe yield.

        Yields:
            The command if it can be used right now.
        """
        if self.screen.check_action(command.action_name(), ()):
            yield command()

    def commands(self) -> CommandHits:
        """Provide the main application commands for the command palette.

        Yields:
            The commands for the command palette.
        """
        yield from self._maybe(Backward)
        yield from self._maybe(BookmarkLocation)
        yield ChangeCommandLineLocation()
        yield ChangeNavigationSide()
        yield ChangeTheme()
        yield from self._maybe(CopyLocationToClipboard)
        yield from self._maybe(CopyMarkdownToClipboard)
        yield from self._maybe(Edit)
        yield from self._maybe(Forward)
        yield Help()
        yield JumpToCommandLine()
        yield JumpToTableOfContents()
        yield JumpToBookmarks()
        yield JumpToHistory()
        yield JumpToLocalBrowser()
        yield Quit()
        yield from self._maybe(Reload)
        yield from self._maybe(SaveCopy)
        yield SearchBookmarks()
        yield ToggleNavigation()


### main.py ends here



"""Provides a command for changing the directory of the local file viewer."""

##############################################################################
# Python imports.
from pathlib import Path

##############################################################################
# Textual imports.
from textual.widget import Widget

##############################################################################
# Local imports.
from ...messages import SetLocalViewRoot
from .base_command import InputCommand


##############################################################################
class ChangeDirectoryCommand(InputCommand):
    """Change the root directory of the local file browser"""

    COMMAND = "`chdir`"
    ALIASES = "`cd`, `dir`, `ls`"
    ARGUMENTS = "`<directory>`"

    @classmethod
    def handle(cls, text: str, for_widget: Widget) -> bool:
        """Handle the command.

        Args:
            text: The text of the command.
            for_widget: The widget to handle the command for.

        Returns:
            `True` if the command was handled; `False` if not.
        """
        command, directory = cls.split_command(text)
        if (
            cls.is_command(command)
            and directory
            and (root := Path(directory).expanduser()).is_dir()
        ):
            for_widget.post_message(SetLocalViewRoot(Path(root).resolve()))
            return True
        return False


### change_directory.py ends here



"""Provides general application commands for the command line."""

##############################################################################
# Python imports.
from functools import partial
from typing import Callable

##############################################################################
# Textual imports.
from textual.message import Message
from textual.widget import Widget

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import Help, Quit

##############################################################################
# Local imports.
from ...commands import (
    JumpToBookmarks,
    JumpToHistory,
    JumpToLocalBrowser,
    JumpToTableOfContents,
)
from ...messages import HandleInput
from .base_command import InputCommand


##############################################################################
class GeneralCommand(InputCommand):
    """Base class for general commands."""

    MESSAGE: Callable[[], Message]
    """The message to send for the command."""

    @classmethod
    def handle(cls, text: str, for_widget: Widget) -> bool:
        """Handle the command.

        Args:
            text: The text of the command.
            for_widget: The widget to handle the command for.

        Returns:
            `True` if the command was handled; `False` if not.
        """
        if cls.is_command(text):
            for_widget.post_message(cls.MESSAGE())
            return True
        return False


##############################################################################
class BookmarksCommand(GeneralCommand):
    """Jump to the bookmarks"""

    COMMAND = "`bookmarks`"
    ALIASES = "`b`, `bm`"
    MESSAGE = JumpToBookmarks


##############################################################################
class ContentsCommand(GeneralCommand):
    """Jump to the table of contents"""

    COMMAND = "`contents`"
    ALIASES = "`c`, `toc`"
    MESSAGE = JumpToTableOfContents


##############################################################################
class HelpCommand(GeneralCommand):
    """Show the help screen"""

    COMMAND = "`help`"
    ALIASES = "`?`"
    MESSAGE = Help


##############################################################################
class HistoryCommand(GeneralCommand):
    """Jump to the browsing history"""

    COMMAND = "`history`"
    ALIASES = "`h`"
    MESSAGE = JumpToHistory


##############################################################################
class LocalCommand(GeneralCommand):
    """Jump to the local file browser"""

    COMMAND = "`local`"
    ALIASES = "`l`"
    MESSAGE = JumpToLocalBrowser


##############################################################################
class QuitCommand(GeneralCommand):
    """Quit the application"""

    COMMAND = "`quit`"
    ALIASES = "`q`"
    MESSAGE = Quit


##############################################################################
class ChangeLogCommand(GeneralCommand):
    """Show Hike's ChangeLog"""

    COMMAND = "`changelog`"
    ALIASES = "`cl`"
    MESSAGE = partial(HandleInput, "github davep hike ChangeLog.md")


##############################################################################
class ReadMeCommand(GeneralCommand):
    """Show Hike's README"""

    COMMAND = "`readme`"
    MESSAGE = partial(HandleInput, "github davep hike")


### general.py ends here



"""Provides the history for the navigation panel."""

##############################################################################
# Python imports.
from pathlib import Path
from typing import cast

##############################################################################
# Rich imports.
from rich.text import Text

##############################################################################
# Textual imports.
from textual import on, work
from textual.binding import Binding
from textual.widgets.option_list import Option

##############################################################################
# Textual-enhanced imports.
from textual_enhanced.dialogs import Confirm
from textual_enhanced.widgets import EnhancedOptionList

##############################################################################
# Local imports.
from ...messages import ClearHistory, OpenFromHistory, RemoveHistoryEntry
from ...types import HikeHistory, HikeLocation


##############################################################################
class Location(Option):
    """A location within the history."""

    def __init__(self, location_id: int, location: HikeLocation) -> None:
        """Initialise the location object.

        Args:
            location_id: The ID of the location within the history.
            location: The location.
        """
        self.location_id = location_id
        """The ID of the location within the history."""
        super().__init__(
            Text.from_markup(
                f":page_facing_up: [bold]{location.name}[/]\n[dim]{location.parent}[/]",
                overflow="ellipsis",
            )
            if isinstance(location, Path)
            else Text.from_markup(
                f":globe_with_meridians: [bold]{Path(location.path).name}[/]"
                f"\n[dim]{Path(location.path).parent}\n{location.host}[/]",
                overflow="ellipsis",
            ),
            id=str(location_id),
        )


##############################################################################
class HistoryView(EnhancedOptionList):
    """The display of history."""

    DEFAULT_CSS = """
    HistoryView {
        height: 1fr;
        border: none;
        &:focus {
            border: none;
        }
    }
    """

    BINDINGS = [
        Binding(
            "delete",
            "remove",
            "Remove",
            show=True,
            tooltip="Remove the current location from history",
        ),
        Binding(
            "backspace",
            "clear",
            "Clear",
            show=True,
            tooltip="Clear all locations from history",
        ),
    ]

    def update(self, history: HikeHistory) -> None:
        """Update the content of the history view.

        Args:
            history: The history to update with.
        """
        self.clear_options().add_options(
            Location(location_id, location)
            for location_id, location in reversed(list(enumerate(history)))
        )
        # If history has been updated, that implies something new has been
        # added to the start; so in that case we jump the highlight to the
        # top.
        if self.option_count:
            self.highlighted = 0

    @on(EnhancedOptionList.OptionHighlighted)
    def visit_from_history(self, message: EnhancedOptionList.OptionHighlighted) -> None:
        """Visit a location from history.

        Args:
            message: The message to say which option was selected.
        """
        message.stop()
        assert isinstance(message.option, Location)
        self.post_message(OpenFromHistory(message.option.location_id))

    def highlight_location(self, location: int) -> None:
        """Highlight an item in the history.

        Args:
            location: The ID of the location to highlight.
        """
        self.highlighted = self.get_option_index(str(location))

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        """Check if an action is possible to perform right now.

        Args:
            action: The action to perform.
            parameters: The parameters of the action.

        Returns:
            `True` if it can perform, `False` or `None` if not.
        """
        if action == "remove":
            return bool(self.option_count) and self.highlighted is not None
        if action == "clear":
            return bool(self.option_count)
        return True

    @work
    async def action_remove(self) -> None:
        """Remove an item of history."""
        if not self.check_action("remove", ()):
            return
        if await self.app.push_screen_wait(
            Confirm(
                "Remove?", "Are you sure you want to remove this location from history?"
            )
        ):
            assert self.highlighted is not None
            location = cast(Location, self.get_option_at_index(self.highlighted))
            self.post_message(RemoveHistoryEntry(location.location_id))

    @work
    async def action_clear(self) -> None:
        """Clear all items from history."""
        if not self.check_action("clear", ()):
            return
        if await self.app.push_screen_wait(
            Confirm(
                "Clear?", "Are you sure you want to clear all locations from history?"
            )
        ):
            self.post_message(ClearHistory())


### history.py ends here



"""Provides a widget for getting input from the user."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from itertools import chain
from typing import Final

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.message import Message
from textual.reactive import var
from textual.suggester import SuggestFromList
from textual.widgets import Input, Label, Rule
from textual.widgets.input import Selection

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import Quit

##############################################################################
# Local imports.
from ...types import CommandHistory
from .base_command import InputCommand
from .change_directory import ChangeDirectoryCommand
from .general import (
    BookmarksCommand,
    ChangeLogCommand,
    ContentsCommand,
    HelpCommand,
    HistoryCommand,
    LocalCommand,
    QuitCommand,
    ReadMeCommand,
)
from .open_directory import OpenDirectoryCommand
from .open_file import OpenFileCommand
from .open_from_forge import (
    OpenFromBitbucket,
    OpenFromCodeberg,
    OpenFromForgeCommand,
    OpenFromGitHub,
    OpenFromGitLab,
)
from .open_url import OpenURLCommand

##############################################################################
COMMANDS: Final[tuple[type[InputCommand], ...]] = (
    # Keep the first three in order. A file match should win over a
    # directory should win over a URL.
    OpenFileCommand,
    OpenDirectoryCommand,
    OpenURLCommand,
    # Once the above are out of the way the order doesn't matter so much.
    BookmarksCommand,
    ChangeDirectoryCommand,
    HelpCommand,
    HistoryCommand,
    LocalCommand,
    ChangeLogCommand,
    OpenFromBitbucket,
    OpenFromCodeberg,
    OpenFromGitHub,
    OpenFromGitLab,
    ContentsCommand,
    ReadMeCommand,
    QuitCommand,
)
"""The commands used for the input."""


##############################################################################
class CommandLine(Vertical):
    """A command line for getting input from the user."""

    DEFAULT_CSS = """
    CommandLine {
        height: 1;

        Label, Input {
            color: $text-muted;
        }

        &:focus-within {
            Label, Input {
                color: $text;
            }
            Label {
                text-style: bold;
            }
        }

        Rule {
            height: 1;
            margin: 0 !important;
            color: $foreground 10%;
            display: none;
        }

        Input, Input:focus {
            border: none;
            padding: 0;
            height: 1fr;
            background: transparent;
        }

        &.--top {
            dock: top;
            height: 2;
            Rule {
                display: block;
            }
        }
    }
    """

    HELP = f"""
    ## Command Line

    Use this command line to enter filenames, directories, URLs or commands. Entering
    a filename or a URL will open that file for viewing; entering a
    directory will open a file opening dialog starting at that location.

    | Command | Aliases | Arguments | Description |
    | --      | --      | --        | --          |
    {'\n    '.join(sorted(command.help_text() for command in COMMANDS))}

    ### ¹Forge support

    The forge-oriented commands listed above accept a number of different
    ways of quickly specifying which file you want to view. Examples include:

    {OpenFromForgeCommand.HELP}
    """

    BINDINGS = [
        ("escape", "request_exit"),
        ("up", "history_previous"),
        ("down", "history_next"),
    ]

    history: var[CommandHistory] = var(CommandHistory)
    """The command line history."""

    dock_top: var[bool] = var(False)
    """Should the input dock to the top of the screen?"""

    @property
    def _history_suggester(self) -> SuggestFromList:
        """A suggester for the history of input.

        If there us no history yet then a list of commands and aliases will
        be used.
        """
        return SuggestFromList(
            [
                # Start off with the history, with the most recently-used
                # commands first so suggestions come from the thing
                # most-recently done.
                *reversed(list(self.history)),
                # Tack known commands on the end; this means that the user
                # will get prompted for commands they've not used yet.
                *chain(*(command.suggestions() for command in COMMANDS)),
            ]
        )

    def compose(self) -> ComposeResult:
        """Compose the content of the widget."""
        with Horizontal():
            yield Label("> ")
            yield Input(
                placeholder="Enter a directory, file, path or command",
                suggester=self._history_suggester,
            )
        yield Rule(line_style="heavy")

    def _watch_history(self) -> None:
        """React to history being updated."""
        if self.is_mounted:
            self.query_one(Input).suggester = self._history_suggester

    def _watch_dock_top(self) -> None:
        """React to being asked to dock input to the top."""
        self.set_class(self.dock_top, "--top")

    @dataclass
    class HistoryUpdated(Message):
        """Message posted when the command history is updated."""

        command_line: CommandLine
        """The command line whose history was updated."""

    def handle_input(self, command: str) -> None:
        """Handle input from the user.

        Args:
            command: The command the user entered.
        """
        if not (command := command.strip()):
            return
        for candidate in COMMANDS:
            if candidate.handle(command, self):
                self.history.add(command)
                self.post_message(self.HistoryUpdated(self))
                self.query_one(Input).value = ""
                self.query_one(Input).suggester = self._history_suggester
                return
        self.notify("Unable to handle that input", title="Error", severity="error")

    @on(Input.Submitted)
    def _handle_input(self, message: Input.Submitted) -> None:
        """Handle input from the user.

        Args:
            message: The message requesting input is handled.
        """
        message.stop()
        self.handle_input(message.value)

    def action_request_exit(self) -> None:
        """Request that the application quits."""
        self.post_message(Quit())

    def action_history_previous(self) -> None:
        """Move backwards through the command line history."""
        if value := self.history.current_item:
            self.query_one(Input).value = value
            self.query_one(Input).selection = Selection(0, len(value))
            self.history.backward()

    def action_history_next(self) -> None:
        """Move forwards through the command line history."""
        if self.history.forward() and (value := self.history.current_item) is not None:
            self.query_one(Input).value = value
            self.query_one(Input).selection = Selection(0, len(value))
        else:
            self.query_one(Input).value = ""


### widget.py ends here



"""Provides application-wide messages."""

##############################################################################
# Local imports.
from .clipboard import CopyToClipboard
from .command_line import HandleInput
from .history import ClearHistory, RemoveHistoryEntry
from .local_view import SetLocalViewRoot
from .opening import OpenFrom, OpenFromForge, OpenFromHistory, OpenLocation

##############################################################################
# Exports.
__all__ = [
    "ClearHistory",
    "CopyToClipboard",
    "HandleInput",
    "OpenFrom",
    "OpenFromForge",
    "OpenFromHistory",
    "OpenLocation",
    "RemoveHistoryEntry",
    "SetLocalViewRoot",
]

### __init__.py ends here



"""Provides code for holding, saving and loading bookmarks."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from functools import total_ordering
from json import dumps, loads
from pathlib import Path
from typing import TypeAlias

##############################################################################
# httpx imports.
from httpx import URL

##############################################################################
# Local imports.
from ..types import HikeLocation
from .locations import data_dir


##############################################################################
@dataclass(frozen=True)
@total_ordering
class Bookmark:
    """A Hike bookmark."""

    title: str
    """The bookmark's title."""

    location: HikeLocation
    """The location of the bookmark."""

    @classmethod
    def from_json(cls, data: dict[str, str]) -> Bookmark:
        """Load a bookmark from some JSON data.

        Args:
            data: The data to load from.

        Returns:
            A fresh instance of a bookmark.
        """
        location_type = data.get("type", "path")
        return cls(
            data.get("title", ""),
            (URL if location_type == "url" else Path)(data.get("location", "")),
        )

    @property
    def as_json(self) -> dict[str, str]:
        """The bookmark in a JSON-friendly format."""
        return {
            "title": self.title,
            "type": "url" if isinstance(self.location, URL) else "path",
            "location": str(self.location),
        }

    def __gt__(self, value: object, /) -> bool:
        if isinstance(value, Bookmark):
            return self.title.casefold() > value.title.casefold()
        raise NotImplementedError

    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, Bookmark):
            return self.title.casefold() == value.title.casefold()
        if isinstance(value, str):
            return self.title.casefold() == value.casefold()
        if isinstance(value, Path):
            return isinstance(self.location, Path) and self.location == value
        if isinstance(value, URL):
            return isinstance(self.location, URL) and self.location == value
        raise NotImplementedError


##############################################################################
Bookmarks: TypeAlias = list[Bookmark]
"""The type for a collection of bookmarks."""


##############################################################################
def bookmarks_file() -> Path:
    """The path of the bookmarks file.

    Returns:
        The path for the bookmarks file.
    """
    return data_dir() / "bookmarks.json"


##############################################################################
def save_bookmarks(bookmarks: Bookmarks) -> None:
    """Save the bookmarks to storage.

    Args:
        bookmarks: The bookmarks to save.
    """
    bookmarks_file().write_text(
        dumps([bookmark.as_json for bookmark in bookmarks], indent=4), encoding="utf-8"
    )


##############################################################################
def load_bookmarks() -> Bookmarks:
    """Load bookmarks from storage.

    Returns:
        The bookmarks.
    """
    return (
        [
            Bookmark.from_json(data)
            for data in loads(bookmarks_file().read_text(encoding="utf-8"))
        ]
        if bookmarks_file().exists()
        else []
    )


### bookmarks.py ends here



"""Commands for the fallback editor."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import (
    ChangeTheme,
    Command,
    CommandHits,
    CommandsProvider,
    Help,
)


##############################################################################
class Save(Command):
    """Save changes back to the document"""

    BINDING_KEY = "f2, ctrl+s"
    SHOW_IN_FOOTER = True


##############################################################################
class Close(Command):
    """Close the editor"""

    BINDING_KEY = "f10"
    SHOW_IN_FOOTER = True


##############################################################################
class EditorCommands(CommandsProvider):
    """Provides the commands for the fallback editor."""

    def commands(self) -> CommandHits:
        """Provide the main application commands for the command palette.

        Yields:
            The commands for the command palette.
        """
        yield ChangeTheme()
        yield Help()
        yield Save()
        yield Close()


### commands.py ends here



"""Provides functions and classes for managing the app's data."""

##############################################################################
# Local imports.
from .bookmarks import Bookmark, Bookmarks, load_bookmarks, save_bookmarks
from .command_history import load_command_history, save_command_history
from .config import (
    Configuration,
    load_configuration,
    save_configuration,
    update_configuration,
)
from .history import load_history, save_history
from .location_types import is_editable, looks_urllike, maybe_markdown

##############################################################################
# Exports.
__all__ = [
    "Bookmark",
    "Bookmarks",
    "Configuration",
    "is_editable",
    "load_bookmarks",
    "load_command_history",
    "load_configuration",
    "load_history",
    "looks_urllike",
    "maybe_markdown",
    "save_bookmarks",
    "save_command_history",
    "save_configuration",
    "save_history",
    "update_configuration",
]

### __init__.py ends here



"""Bookmark search and visit commands for the command palette."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import CommandHit, CommandHits, CommandsProvider

##############################################################################
# Local imports.
from ..data import Bookmarks
from ..messages import OpenLocation


##############################################################################
class BookmarkCommands(CommandsProvider):
    """A command palette provider related to bookmarks."""

    bookmarks: Bookmarks = Bookmarks()
    """The bookmarks."""

    @classmethod
    def prompt(cls) -> str:
        """The prompt for the command provider."""
        return "Search bookmarks..."

    def commands(self) -> CommandHits:
        """Provide the bookmark-based command data for the command palette.

        Yields:
            The commands for the command palette.
        """
        for bookmark in sorted(self.bookmarks):
            yield CommandHit(
                f"Visit {bookmark.title}",
                f"{bookmark.location}",
                OpenLocation(bookmark.location),
            )


### bookmarks.py ends here



"""Provides the base command class."""

##############################################################################
# Textual imports.
from textual.widget import Widget


##############################################################################
class InputCommand:
    """Base class for input commands."""

    COMMAND = ""
    """The command for the help."""

    ALIASES = ""
    """Any aliases for the command for the help."""

    ARGUMENTS = ""
    """Any arguments for the command for the help."""

    @classmethod
    def help_text(cls) -> str:
        """Get the help text for the command.

        Returns:
            The help text formatted as a Markdown table row.
        """
        return f"| {cls.COMMAND} | {cls.ALIASES} | {cls.ARGUMENTS} | {cls.__doc__} |"

    @classmethod
    def handle(cls, text: str, for_widget: Widget) -> bool:
        """Handle the command.

        Args:
            text: The text of the command.
            for_widget: The widget to handle the command for.

        Returns:
            `True` if the command was handled; `False` if not.
        """
        return False

    @staticmethod
    def split_command(text: str) -> tuple[str, str]:
        """Split the command for further testing.

        Args:
            text: The text of the command.

        Returns:
            The command and its tail.
        """
        command, _, tail = text.strip().partition(" ")
        return command.strip(), tail.strip()

    @classmethod
    def suggestions(cls) -> tuple[str, ...]:
        """The suggested command matches for this command.

        Returns:
            A tuple of suggestions for the command.

        Notes:
            Any commands that look like value placeholders will be filtered
            out.
        """
        # Build up all the possible matches. These are built from the main
        # command and also the aliases. By convention the code will often
        # use `code` fences for commands, and the aliases will be a comma
        # list, so we clean that up as we go...
        return tuple(
            candidate.strip().removeprefix("`").removesuffix("`")
            for candidate in (cls.COMMAND, *cls.ALIASES.split(","))
            # Note that we filter out anything that looks like `<this>`.
            if not candidate.startswith("`<")
        )

    @classmethod
    def is_command(cls, command: str) -> bool:
        """Does the given command appear to be a match?

        Args:
            command: The command to test.

        Returns:
            `True` if the given command seems to be a match, `False` if not.
        """
        return command.strip().lower() in cls.suggestions()


### base_command.py ends here
